/*  
 *  main_basic_ft_cor.cpp: Ray-casting based volume rendering 
 *                  for JAI 2013/CSBC mini-course (OpenGL 1.0)
 *
 *  Copyright (C) 2013  Wu Shin-Ting, FEEC, Unicamp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

//http://www.khronos.org/files/opengl-quick-reference-card.pdf
#ifdef WIN32
#include <windows.h>
#include <GL/glew.h>   
#else
#include <getopt.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <cstdio>
#include <cstring>
#include <string>
#include <iostream>
#include <stdio.h>     /* for printf */
#include <stdint.h>
#include <stdlib.h>    /* for exit */
#include <string.h>
#include <limits.h>
#include <math.h>
#include <GL/glut.h>    // Header File For The GLUT Library

#include "importDicom.h"
#include "transferFunction.h"

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

GLfloat spin_x=0.0, spin_y=0.0, spin_z=0.0;
GLfloat nVolSideZ=1.0;

Import load;
TransferFunction tfunc;

GLuint texID[2];
Import::ImgFormat data;
GLfloat nVolSides[3], volCenter[3];
GLuint win, VIEWPORT_WIDTH=500, VIEWPORT_HEIGHT=500;

void transfereTexturas() {
  unsigned char *tf;
  int volSize, intensidade;
  unsigned char *texbuffer;

  // Gera nome de 1 texturas
  glGenTextures(1, texID);

  // Define a unidade de alinhamento nos acessos
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  // Constroi a funcao de transferencia (0=logaritmica; 1=linear; 2=quadratico; 3 cubico)
  int dim;
  tfunc.GetRGBTF (1, data.umin, data.umax, &dim, &tf);

  // Mapeia as intensidades em propriedades opticas (niveis de cinza)
  volSize = data.dims[0]*data.dims[1]*data.dims[2];
  texbuffer = new unsigned char[volSize*3];
  memset(texbuffer, 0x00, volSize*3);

  int di, iz, iy, ix, aux; 
  for (di=0, iz = 0; iz < data.dims[2]; iz++) {
    for (iy =0; iy < data.dims[1]; iy++) {
      for (ix =0; ix < data.dims[0]; ix++) {
	intensidade = (int)((reinterpret_cast<unsigned short*>(data.buffer))[iz*data.dims[0]*data.dims[1]+iy*data.dims[0]+ix]);
	aux = intensidade/7;
	texbuffer[di] = (unsigned char)(tf[aux*3]);
	texbuffer[di+1] = (unsigned char)(tf[aux*3+1]);
	texbuffer[di+2] = (unsigned char)(tf[aux*3+2]);
	di += 3;
      }
    }
  }

  // Carrega o volume em niveis de cinza como textura 3D
  glBindTexture(GL_TEXTURE_3D, texID[0]);
  glTexImage3D(GL_TEXTURE_3D, 0, GL_RGB, data.dims[0],
  	       data.dims[1], data.dims[2], 0, GL_RGB,
  	       GL_UNSIGNED_BYTE, texbuffer);

  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

  // Libera memoria
  delete []texbuffer;
}

void drawQuads() 
{
  glBegin(GL_QUADS);
  // Coordenadas espaciais dos vértices e as respectivas coordenadas de textura
  // Volume de textura: [1.0,1.0,1.0]
  // Face de fundo
  glTexCoord3f(0.0, 0.0, 0.0);
  glVertex3f(0.0, 0.0, 0.0);
  glTexCoord3f(0.0, 1.0, 0.0);
  glVertex3f(0.0, nVolSides[1], 0.0);
  glTexCoord3f(1.0, 1.0, 0.0);
  glVertex3f(nVolSides[0], nVolSides[1], 0.0);
  glTexCoord3f(1.0, 0.0, 0.0);
  glVertex3f(nVolSides[0], 0.0, 0.0);

  // Face frontal
  glTexCoord3f(0.0, 0.0, nVolSideZ);
  glVertex3f(0.0, 0.0, nVolSideZ*nVolSides[2]);
  glTexCoord3f(1.0, 0.0, nVolSideZ);
  glVertex3f(nVolSides[0], 0.0, nVolSideZ*nVolSides[2]);
  glTexCoord3f(1.0, 1.0, nVolSideZ);
  glVertex3f(nVolSides[0], nVolSides[1], nVolSideZ*nVolSides[2]);
  glTexCoord3f(0.0, 1.0, nVolSideZ);
  glVertex3f(0.0, nVolSides[1], nVolSideZ*nVolSides[2]);

  // Face de topo
  glTexCoord3f(0.0, 1.0, 0.0);
  glVertex3f(0.0, nVolSides[1], 0.0);
  glTexCoord3f(0.0, 1.0, nVolSideZ);
  glVertex3f(0.0, nVolSides[1], nVolSideZ*nVolSides[2]);
  glTexCoord3f(1.0, 1.0, nVolSideZ);
  glVertex3f(nVolSides[0], nVolSides[1], nVolSideZ*nVolSides[2]);
  glTexCoord3f(1.0, 1.0, 0.0);
  glVertex3f(nVolSides[0], nVolSides[1], 0.0);

  // Face de apoio
  glTexCoord3f(0.0, 0.0, 0.0);
  glVertex3f(0.0, 0.0, 0.0);
  glTexCoord3f(1.0, 0.0, 0.0);
  glVertex3f(nVolSides[0], 0.0, 0.0);
  glTexCoord3f(1.0, 0.0, nVolSideZ);
  glVertex3f(nVolSides[0], 0.0, nVolSideZ*nVolSides[2]);
  glTexCoord3f(0.0, 0.0, nVolSideZ);
  glVertex3f(0.0, 0.0, nVolSideZ*nVolSides[2]);

  // Face esquerda
  glTexCoord3f(0.0, 0.0, 0.0);
  glVertex3f(0.0, 0.0, 0.0);
  glTexCoord3f(0.0, 0.0, nVolSideZ);
  glVertex3f(0.0, 0.0, nVolSideZ*nVolSides[2]);
  glTexCoord3f(0.0, 1.0, nVolSideZ);
  glVertex3f(0.0, nVolSides[1], nVolSideZ*nVolSides[2]);
  glTexCoord3f(0.0, 1.0, 0.0);
  glVertex3f(0.0, nVolSides[1], 0.0);

  // Face direita
  glTexCoord3f(1.0, 0.0, 0.0);
  glVertex3f(nVolSides[0], 0.0, 0.0);
  glTexCoord3f(1.0, 1.0, 0.0);
  glVertex3f(nVolSides[0], nVolSides[1], 0.0);
  glTexCoord3f(1.0, 1.0, nVolSideZ);
  glVertex3f(nVolSides[0], nVolSides[1], nVolSideZ*nVolSides[2]);
  glTexCoord3f(1.0, 0.0, nVolSideZ);
  glVertex3f(nVolSides[0], 0.0, nVolSideZ*nVolSides[2]);
  glEnd();
}

void desenheProxyGeometry() 
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Seta os parametros de transformacao geometrica
  glLoadIdentity();
  glTranslatef(0., 0., -5.);
  glTranslatef(volCenter[0], volCenter[1], volCenter[2]);
  glRotatef(spin_x, 1.0, 0.0, 0.0);
  glRotatef(spin_y, 0.0, 1.0, 0.0);
  glRotatef(spin_z, 0.0, 0.0, 1.0);
  glTranslatef(-volCenter[0], -volCenter[1], -volCenter[2]);

  drawQuads();

  //Esvazia os buffers  
  glFlush();
}

void init()
{
  // Define a cor em cinza (0.5,0.5,0.5) como cor de "CLEAR" */
  glClearColor (0.5, 0.5, 0.5, 1.0);

  // Transfere os dados como textura para GPU (servidor)
  transfereTexturas();

  // Normaliza as coordenadas em relação ao maior lado do volume
  float volSides[3], maxVolSide;
  int i;

  for (i = 0; i < 3; i++) {
    volSides[i] = data.dims[i] * data.space[i];
  }

  maxVolSide = MAX(MAX(volSides[0], volSides[1]), volSides[2]);

  for (i = 0; i < 3; i++) {
    nVolSides[i] = volSides[i]/maxVolSide;
    volCenter[i] = nVolSides[i]/2.0;
  }

  // Habilita cull_face 
  glEnable(GL_CULL_FACE);

  // Seta modo de renderizacao
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  /* habilita a texturização */
  glEnable(GL_TEXTURE_3D);
}

void reshape(int x, int y)
{
    VIEWPORT_WIDTH = x; VIEWPORT_HEIGHT = y;

    glViewport(0, 0, VIEWPORT_WIDTH, VIEWPORT_HEIGHT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-0.5, 1.5, -0.5, 1.5, 0.0, 8.0); 

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void keyboard(unsigned char key, int x, int y)
{
  //float v;

  switch (key) {
  case 'x':
  case 'X':
    /* incrementa +5.0 no ângulo de rotação em torno do eixo x */
    spin_x += 5.0;
    break;
  case 'y':
  case 'Y':
    /* incrementa +5.0 no ângulo de rotação em torno do eixo y */
    spin_y += 5.0;
    break;
  case 'z':
  case 'Z':
   /* incrementa +5.0 no ângulo de rotação em torno do eixo z */
    spin_z += 5.0;
    break;
  case 'r':
  case 'R':
    /* Reseta */
    spin_x = spin_y = spin_z = 0.0;
    nVolSideZ = 1.0;
    break;
  case '+':
    /* retrocede o plano de corte */
    if (nVolSideZ + 0.05 < 1.0) nVolSideZ += 0.05;
    else nVolSideZ = 1.0;
    break;
  case '-':
    /* avança o plano de corte */
    if (nVolSideZ - 0.05 > 0.0) nVolSideZ -= 0.05;
    else nVolSideZ = 0.0;
    break;
  case 27:
    /* libera memoria no cliente */
    delete [] data.buffer;
    /* libera memoria no servidor */
    glDeleteTextures (1, texID);
    exit(0);
    return;
  }
  /* Gerar o evento "Display" */
  glutPostRedisplay();
}

int main(int argc, char *argv[])
{
  int c;
  std::string filename;

  /* Carrega a imagem 3D DICOM */
  if (argc == 2) {
    filename=argv[1];
    data.umax = -pow(2,30);
    data.umin = pow(2,30);
    if (!load.DICOMImage (filename, &data)) {
      std::cout << filename << " cannot be imported!" << std::endl;
      return 0;
    }
    std::cout << "Faixa de intensidades: " << data.umin << " - " << data.umax << std::endl; 
  } else {
    std::cout << "Comando: ./jai13_basic_ft_cor <caminho completo do volume de dados>" << std::endl;
    return 0;
  }

  /* Inicializa OpenGL */
  glutInit(&argc, argv);

  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_SINGLE);

  /* Seta contexto GL */
  win = glutCreateWindow("Volume de Dados");

  // Verifca a versao de OpenGL: glxinfo | grep "OpenGL version"
  std::cout << "OpenGL version supported by this platform: " << glGetString(GL_VERSION) << std::endl;

  /* Outras inicializacoes */
#ifdef WIN32
  glewInit();
#endif
  init();

  /* Configura a janela */
  glutPositionWindow(10, 10);
  glutReshapeWindow(VIEWPORT_WIDTH, VIEWPORT_HEIGHT);

  /* Define os tratadores de eventos */
  glutReshapeFunc(reshape);
  glutDisplayFunc(desenheProxyGeometry);
  glutKeyboardFunc(keyboard);

  /* Espera por eventos */
  glutMainLoop();
 
  return 0;
}
