/*  
 *  main_glsl_dvr_int.cpp: Ray-casting based volume rendering for 
 *                 JAI 2013/CSBC mini-course (OpenGL 2.0 - GLSL)
 *
 *  Copyright (C) 2013  Wu Shin-Ting, FEEC, Unicamp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
//http://www.khronos.org/files/opengl-quick-reference-card.pdf
#ifdef WIN32
#include <windows.h>
#include <GL/glew.h>    
#else
#include <getopt.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <cstdio>
#include <cstring>
#include <string>
#include <iostream>
#include <stdio.h>     /* for printf */
#include <stdint.h>
#include <stdlib.h>    /* for exit */
#include <string.h>
#include <limits.h>
#include <math.h>
#include <GL/glut.h>    // Header File For The GLUT Library

#include "importDicom.h"
#include "transferFunction.h"
#include "equalize.h"
#include "shaders.h"

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

float vertices[][3] = {  // coordenadas espaciais dos vértices
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}
};  

// Volume de textura: [1.0,1.0,1.0]
float texCoords0[][3] = {   // respectivas coordenadas de textura 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 1.0f, 0.0f}, 
  {1.0f, 1.0f, 0.0f},
  {1.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 1.0f},
  {1.0f, 0.0f, 1.0f}, 
  {1.0f, 1.0f, 1.0f},
  {0.0f, 1.0f, 1.0f},

  {0.0f, 1.0f, 0.0f},
  {0.0f, 1.0f, 1.0f}, 
  {1.0f, 1.0f, 1.0f},
  {1.0f, 1.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {1.0f, 0.0f, 0.0f}, 
  {1.0f, 0.0f, 1.0f},
  {0.0f, 0.0f, 1.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 1.0f}, 
  {0.0f, 1.0f, 1.0f},
  {0.0f, 1.0f, 0.0f},

  {1.0f, 0.0f, 0.0f},
  {1.0f, 1.0f, 0.0f}, 
  {1.0f, 1.0f, 1.0f},
  {1.0f, 0.0f, 1.0f}
};

GLubyte quadIndices[] = {
    0,1,2,3,
    4,5,6,7,
    8,9,10,11,
    12,13,14,15,
    16,17,18,19,
    20,21,22,23
};

GLfloat spin_x=0.0, spin_y=0.0, spin_z=0.0;
GLfloat scale=1.0;
GLfloat nVolSideZ=0.5, alpha=0.02, passo=0.005, noise=0.25;
static GLubyte g_bButton1Down = 0, g_bButton2Down = 0;
int prev_winX, prev_winY;
float prev_Pos[3];
double prev_Qt[]={0.,0.,0.,1.};
float accumMat[] = {1.,0.,0.,0.,0.,1.,0.,0.,0.,0.,1.,0.,0.,0.,0.,1.};
float prev_winZ=0.5;
int H;
GLdouble cursor[3];
float eixo;

Import load;
TransferFunction tfunc;
Equalize eq;
Shaders sh;

const char vertexFileName[] = "shaders/dvr.vert";
const char cursorFileName[] = "shaders/cursor.vert";
const char fragmentFileName[] = "shaders/dvr_int.frag";
const char fragmentDepthFileName[] = "shaders/depth.frag";
const char fragmentCursorFileName[] = "shaders/cursor.frag";

GLuint texID[3];
Import::ImgFormat data;
GLfloat nVolSides[4], volCenter[4];
GLuint win, VIEWPORT_HEIGHT=700, VIEWPORT_WIDTH=700;
GLuint shader_dvr, shader_depthmap, shader_cursor;
GLuint vao=0, vbo[3]={0,0,0};
GLuint fbo = 0;
float *bufferTex2D, *depthBuffer3D;

int transfereTexturas() {
  unsigned char *tf;
  unsigned int *mapa;

  // Gera nomes de 3 texturas: 2 para renderizacao e 1 para mapa de textura
  glGenTextures(3, texID);

  // Define a unidade de alinhamento nos acessos
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  // Equaliza o histograma para aumentar a escala dinamica das intensidades
  // como uma forma de contornar a perda de resolucao durante normalizacao 
  // feita pelas unidades de textura
  eq.EqualizeHistogram (data.dims[0], data.dims[1], data.dims[2], 
  			(reinterpret_cast<unsigned short*>(data.buffer)), 
                        data.nbitsalloc, data.umax, &mapa);

  int intensidade, di, iz, iy, ix, volSize; 
  unsigned short *texbuffer;

  volSize = data.dims[0]*data.dims[1]*data.dims[2];
  texbuffer = new unsigned short[volSize];
  memset(texbuffer, 0x00, volSize*sizeof(unsigned short));

  for (di=0, iz = 0; iz < data.dims[2]; iz++) {
    for (iy =0; iy < data.dims[1]; iy++) {
      for (ix =0; ix < data.dims[0]; ix++) {
  	intensidade = (int)((reinterpret_cast<unsigned short*>(data.buffer))[iz*data.dims[0]*data.dims[1]+iy*data.dims[0]+ix]);
  	texbuffer[di++] = (unsigned short)(mapa[intensidade]);
      }
    }
  }

  // Carrega o volume equalizado como textura 3D
  glBindTexture(GL_TEXTURE_3D, texID[0]);
  glTexImage3D(GL_TEXTURE_3D, 0, GL_LUMINANCE, data.dims[0],
  	       data.dims[1], data.dims[2], 0, GL_LUMINANCE,
  	       GL_UNSIGNED_SHORT, texbuffer);

  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

  delete [] texbuffer;

  // Constroi a funcao de transferencia para uma faixa de [umin,umax]
  // valores validos no dominio de 2^{nbitsalloc}
  int dim;
  tfunc.GetGrayScaleTF (4, 0, pow(2,data.nbitsalloc), &dim, &tf, data.nbitsalloc);

  // Carrega a funcao de transferencia como textura 1D
  glBindTexture(GL_TEXTURE_1D, texID[1]);
  glTexImage1D(GL_TEXTURE_1D, 0, GL_LUMINANCE, dim, 0, GL_LUMINANCE,
  	       GL_UNSIGNED_BYTE, tf);

  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

  delete []tf;

  return 1;  
}

void crieVertexArray() {
  float volSides[3], maxVolSide;
  int i;

  // Coordenadas espaciais dos vértices e as respectivas coordenadas de textura
  /* Face de fundo */
  vertices[1][1] = nVolSides[1]; 
  vertices[2][0] = nVolSides[0]; vertices[2][1] = nVolSides[1]; 
  vertices[3][0] = nVolSides[0]; 

  /* Face frontal */
  vertices[4][2] = nVolSides[2];
  vertices[5][0] = nVolSides[0]; vertices[5][2] = nVolSides[2];
  vertices[6][0] = nVolSides[0]; vertices[6][1] = nVolSides[1]; 
  vertices[6][2] = nVolSides[2];
  vertices[7][1] = nVolSides[1]; vertices[7][2] = nVolSides[2];

  /* Face de topo */
  vertices[8][1] = nVolSides[1]; 
  vertices[9][1] = nVolSides[1]; vertices[9][2] = nVolSides[2];
  vertices[10][0] = nVolSides[0]; vertices[10][1] = nVolSides[1]; 
  vertices[10][2] = nVolSides[2];
  vertices[11][0] = nVolSides[0]; vertices[11][1] = nVolSides[1]; 

  /* Face de apoio */
  vertices[13][0] = nVolSides[0]; 
  vertices[14][0] = nVolSides[0]; vertices[14][2] = nVolSides[2];
  vertices[15][2] = nVolSides[2];

  /* Face esquerda */
  vertices[17][2] = nVolSides[2];
  vertices[18][1] = nVolSides[1]; vertices[18][2] = nVolSides[2];
  vertices[19][1] = nVolSides[1];

  /* Face direita */
  vertices[20][0] = nVolSides[0]; 
  vertices[21][0] = nVolSides[0]; vertices[21][1] = nVolSides[1];
  vertices[22][0] = nVolSides[0]; vertices[22][1] = nVolSides[1]; 
  vertices[22][2] = nVolSides[2];
  vertices[23][0] = nVolSides[0]; vertices[23][2] = nVolSides[2];

  glGenVertexArrays(1, &vao); // gerar o nome para vertex array object
  glBindVertexArray(vao);     // <vao> passa a ser objeto corrente

  glGenBuffers(3, vbo);       // gerar nome para os 3 buffer objects
  
  glBindBuffer(GL_ARRAY_BUFFER, vbo[0]); // associar vbo[0] a <vao>
  glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW); // transferir seus dados para GPU
  glBindBuffer(GL_ARRAY_BUFFER, vbo[1]); // associar vbo[1] a <vao>
  glBufferData(GL_ARRAY_BUFFER, sizeof(texCoords0), texCoords0, GL_STATIC_DRAW);
  
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbo[2]); // associar vbo[2] a <vao>
  glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(quadIndices), quadIndices, GL_STATIC_DRAW);
  glBindVertexArray(0);       // disassociar <vao> como objeto corrente
}

void crieFBO () {
  // Setup para offscreen FB
  // Obter o número máximo de "color attachement points"
  int maxColorBuffers;

  glGetIntegerv( GL_MAX_COLOR_ATTACHMENTS, &maxColorBuffers );

  // alocar 2 áreas de textura do tamanho da janela sem inicialização:
  // color buffer (no mínimo, 1) e depth buffer
  glBindTexture(GL_TEXTURE_RECTANGLE, texID[2]);
  glTexParameterf(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameterf(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexImage2D(GL_TEXTURE_RECTANGLE, 0, GL_RGB,
	       VIEWPORT_WIDTH, VIEWPORT_HEIGHT, 0, GL_RGB,
	       GL_FLOAT, 0);

  // associar o espaço a um framebuffer object
  glGenFramebuffers(1, &fbo);
  glBindFramebuffer(GL_FRAMEBUFFER, fbo);
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
			 GL_TEXTURE_RECTANGLE, texID[2], 0);

  glBindFramebuffer(GL_FRAMEBUFFER, 0); // retornar para o buffer convencional
}

void instaleShaders () {
  // carregue, compile e ligue shaders
  shader_dvr = sh.carregueShaders(vertexFileName, fragmentFileName);   

  shader_depthmap = sh.carregueShaders(vertexFileName, fragmentDepthFileName); 

  shader_cursor = sh.carregueShaders(vertexFileName, fragmentCursorFileName);  
}

void inverseMatrix(GLfloat *mat1, GLfloat *mat2) {
  float s0, s1, s2, s3, s4, s5, c0, c1, c2, c3, c4, c5;
  float det, invDet;

  s0 = mat1[0]*mat1[5] - mat1[1]*mat1[4];
  s1 = mat1[0]*mat1[9] - mat1[1]*mat1[8];
  s2 = mat1[0]*mat1[13] - mat1[1]*mat1[12];
  s3 = mat1[4]*mat1[9] - mat1[5]*mat1[8];
  s4 = mat1[4]*mat1[13] - mat1[5]*mat1[12];
  s5 = mat1[8]*mat1[13] - mat1[9]*mat1[12];

  c5 = mat1[10]*mat1[15] - mat1[11]*mat1[14];
  c4 = mat1[6]*mat1[15] - mat1[7]*mat1[14];
  c3 = mat1[6]*mat1[11] - mat1[7]*mat1[10];
  c2 = mat1[2]*mat1[15] - mat1[3]*mat1[14];
  c1 = mat1[2]*mat1[11] - mat1[3]*mat1[10];
  c0 = mat1[2]*mat1[7] - mat1[3]*mat1[6];

  det = s0 * c5 - s1 * c4 + s2 * c3 + s3 * c2 - s4 * c1 + s5 * c0;

  invDet = 1.0f/det;

  mat2[0] = (mat1[5]*c5 - mat1[9]*c4 + mat1[13]*c3) * invDet;
  mat2[4] = (-mat1[4]*c5 + mat1[8]*c4 - mat1[12]*c3) * invDet;
  mat2[8] = (mat1[7]*s5 - mat1[11]*s4 + mat1[15]*s3) * invDet;
  mat2[12] = (-mat1[6]*s5 + mat1[10]*s4 - mat1[14]*s3) * invDet;

  mat2[1] = (-mat1[1]*c5 + mat1[9]*c2 - mat1[13]*c1) * invDet;
  mat2[5] = (mat1[0]*c5 - mat1[8]*c2 + mat1[12]*c1) * invDet;
  mat2[9] = (-mat1[3]*s5 + mat1[11]*s2 - mat1[15]*s1) * invDet;
  mat2[13] = (mat1[2]*s5 - mat1[10]*s2 + mat1[14]*s1) * invDet;

  mat2[2] = (mat1[1]*c4 - mat1[5]*c2 + mat1[13]*c0) * invDet;
  mat2[6] = (-mat1[0]*c4 + mat1[4]*c2 - mat1[12]*c0) * invDet;
  mat2[10] = (mat1[3]*s4 - mat1[7]*s2 + mat1[15]*s0) * invDet;
  mat2[14] = (-mat1[2]*s4 + mat1[6]*s2 - mat1[14]*s0) * invDet;

  mat2[3] = (-mat1[1]*c3 + mat1[5]*c1 - mat1[9]*c0) * invDet;
  mat2[7] = (mat1[0]*c3 - mat1[4]*c1 + mat1[8]*c0) * invDet;
  mat2[11] = (-mat1[3]*s3 + mat1[7]*s1 - mat1[11]*s0) * invDet;
  mat2[15] = (mat1[2]*s3 - mat1[6]*s1 + mat1[10]*s0) * invDet;
}

void atribuaValores(GLuint shader) {
  GLfloat modelview[16], projection[16], inv_modelview[16];

    glGetFloatv(GL_MODELVIEW_MATRIX, modelview);   // ler a matriz MODELVIEW
    glGetFloatv(GL_PROJECTION_MATRIX, projection); // ler a matriz PROJECTION
    inverseMatrix (modelview, inv_modelview);       // inverter a matriz MODELVIEW
  
    glUniformMatrix4fv(glGetUniformLocation(shader, "viewMatrix"), 1, GL_FALSE, modelview);// passar os valores
    glUniformMatrix4fv(glGetUniformLocation(shader, "projMatrix"), 1, GL_FALSE, projection);
    glUniformMatrix4fv(glGetUniformLocation(shader, "viewMatrixInv"), 1, GL_FALSE, inv_modelview);

  if (shader == shader_cursor) {
    glUniform4f(glGetUniformLocation(shader,"iparam"), eixo, 0, 0, 0);
  } else {
    glUniform4f(glGetUniformLocation(shader,"fparam"), passo, nVolSideZ, alpha, noise);   // passar o passo, o valor z de corte, a opacidade e o limiar
    glUniform4fv(glGetUniformLocation(shader,"scaleFactors"), 1, nVolSides);

    glActiveTexture(GL_TEXTURE0);               //ativa unidade de textura 1
    glEnable(GL_TEXTURE_3D);
    glBindTexture(GL_TEXTURE_3D, texID[0]);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glUniform1i(glGetUniformLocation(shader,"VOLUME"),0);           //passar unidade da textura 3D
    glActiveTexture(GL_TEXTURE0+1);           //ativa unidade de textura 2
    glEnable(GL_TEXTURE_1D);
    glBindTexture(GL_TEXTURE_1D, texID[1]);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glUniform1i(glGetUniformLocation(shader,"TRANSFERFUNCTION"),1);
  }
  if (shader == shader_dvr || shader == shader_cursor) {
    glActiveTexture(GL_TEXTURE0+2);           //ativa unidade de textura 2
    glEnable(GL_TEXTURE_RECTANGLE);
    glBindTexture(GL_TEXTURE_RECTANGLE, texID[2]);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glUniform1i(glGetUniformLocation(shader,"DEPTHMAP"),2);           
  }
}

void desenheCursor()
{
  // Habilita shader de depth map
  glLineWidth (2.0);

  glUseProgram (shader_cursor);
  // Passa os valores dos parametros
  eixo = 0.;
  atribuaValores(shader_cursor);

  glBegin(GL_LINES);
  glVertex3f(cursor[0]-0.15, cursor[1], cursor[2]);
  glVertex3f(cursor[0]+0.15, cursor[1], cursor[2]);
  glEnd();
  
  eixo = 1.;
  atribuaValores(shader_cursor);
  glBegin(GL_LINES);
  glVertex3f(cursor[0], cursor[1]-0.15, cursor[2]);
  glVertex3f(cursor[0], cursor[1]+0.15, cursor[2]);
  glEnd();
  
  eixo = 2.;
  atribuaValores(shader_cursor);
  glBegin(GL_LINES);
  glVertex3f(cursor[0], cursor[1], cursor[2]-0.15);
  glVertex3f(cursor[0], cursor[1], cursor[2]+0.15);
  glEnd();

  // Desabilita os shaders
  glUseProgram(0);
}

void desenheMapaProfundidade()
{
  //======== Primeiro shader
  // Seta como saida o ponto 0 do objeto framebuffer  
  glBindFramebuffer(GL_FRAMEBUFFER, fbo);
  GLenum buffers[1] = {GL_COLOR_ATTACHMENT0};
  
  glDrawBuffers(1, buffers);

  // Assegura que todos os fragmentos tenham valores >= 1.0
  glClearColor (1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);

  // Habilita shader de depth map
  glUseProgram (shader_depthmap);

  // Passa os valores dos parametros
  atribuaValores(shader_depthmap);

  // Associa ponteiros dos buffers ao contexto GL
  glBindVertexArray(vao); // vincular a <vao>
  glBindBuffer(GL_ARRAY_BUFFER, vbo[0]); 
  glEnableVertexAttribArray(glGetAttribLocation(shader_depthmap,"position")); 
  glVertexAttribPointer(glGetAttribLocation(shader_depthmap,"position"), 3, GL_FLOAT, GL_FALSE, 0, (const GLvoid *)0);
  glBindBuffer(GL_ARRAY_BUFFER, vbo[1]); 
  glEnableVertexAttribArray(glGetAttribLocation(shader_depthmap,"texCoord0"));    glVertexAttribPointer(glGetAttribLocation(shader_depthmap,"texCoord0"), 3, GL_FLOAT, GL_FALSE, 0, (const GLvoid *)0);

  // renderizar 6 facetas com 4 vértices em cada uma
  // cujos atributos estão associados a <vao>
  glDrawElements(GL_QUADS, 24, GL_UNSIGNED_BYTE, (const GLvoid *)0);

  // Desassociar vao do contexto GL
  glBindVertexArray(0);

  glUseProgram(0);

  //Recuperar os dados do framebuffer
  glReadBuffer(GL_COLOR_ATTACHMENT0);
  glReadPixels(0, 0, VIEWPORT_WIDTH, VIEWPORT_HEIGHT, GL_RGB, GL_FLOAT,
	       bufferTex2D);

  // Retornar para "window-system-provided framebuffer"
  glBindFramebuffer(GL_FRAMEBUFFER, 0);

  //Gerar o mapa de profundidade
  //Armzenamento por linha com (0,0) no canto esquerdo inferior
  for (int i=0; i < VIEWPORT_WIDTH*VIEWPORT_HEIGHT; i++)
    depthBuffer3D[i] = bufferTex2D[i*3];
}
  
void desenheProxyGeometry() 
{
  //===== Segundo shader  
  glClearColor (0.5, 0.5, 0.5, 1.0);
  glClear(GL_COLOR_BUFFER_BIT);

  // Habilita shader de DVR
  glUseProgram (shader_dvr);

  // Passa os valores dos parametros
  atribuaValores(shader_dvr);

  // Associa ponteiros dos buffers ao contexto GL
  glBindVertexArray(vao); // vincular a <vao>
  glBindBuffer(GL_ARRAY_BUFFER, vbo[0]); 
  glEnableVertexAttribArray(glGetAttribLocation(shader_dvr,"position")); 
  glVertexAttribPointer(glGetAttribLocation(shader_dvr,"position"), 3, GL_FLOAT, GL_FALSE, 0, (const GLvoid *)0);
  glBindBuffer(GL_ARRAY_BUFFER, vbo[1]); 
  glEnableVertexAttribArray(glGetAttribLocation(shader_dvr,"texCoord0"));    
  glVertexAttribPointer(glGetAttribLocation(shader_dvr,"texCoord0"), 3, GL_FLOAT, GL_FALSE, 0, (const GLvoid *)0);

  // renderizar 6 facetas com 4 vértices em cada uma
  // cujos atributos estão associados a <vao>
  glDrawElements(GL_QUADS, 24, GL_UNSIGNED_BYTE, (const GLvoid *)0);

  // Desassociar vao do contexto GL
  glBindVertexArray(0);
  
  // Desabilita os shaders
  glUseProgram(0);
}

void desenhe() {
  // Seta os parametros de transformacao geometrica
  glLoadIdentity();
  glTranslatef (0.0, 0.0, -5.0);
  glTranslatef(volCenter[0], volCenter[1], volCenter[2]);
  glMultMatrixf(accumMat);
  // glRotatef(spin_x, 1.0, 0.0, 0.0);
  // glRotatef(spin_y, 0.0, 1.0, 0.0);
  // glRotatef(spin_z, 0.0, 0.0, 1.0);
  glScalef(scale, scale, scale);
  glTranslatef(-volCenter[0], -volCenter[1], -volCenter[2]);

  glDisable(GL_DEPTH_TEST);

  desenheMapaProfundidade();
  desenheProxyGeometry();
  desenheCursor();

  glEnable(GL_DEPTH_TEST);
  glutSwapBuffers();
}

void init()
{
  // Define a cor em cinza (0.5,0.5,0.5) como cor de "CLEAR" */
  glClearColor (0.5, 0.5, 0.5, 1.0);

  // Transfere os dados como textura para GPU (servidor)
  transfereTexturas();

  // Normaliza as dimensoes do volume em relação ao sue maior lado 
  float volSides[3], maxVolSide;
  int i;

  for (i = 0; i < 3; i++) {
    volSides[i] = data.dims[i] * data.space[i];
  }

  maxVolSide = MAX(MAX(volSides[0], volSides[1]), volSides[2]);

  for (i = 0; i < 3; i++) {
    nVolSides[i] = volSides[i]/maxVolSide;
    volCenter[i] = nVolSides[i]/2.0;
  }

  nVolSides[3] = volCenter[3] = 1.0;

  // Cria vertex buffer no servidor
  crieVertexArray();

  // Cria fbo para armazenar as coordenadas de profundidade
  crieFBO();

  bufferTex2D = new float[VIEWPORT_WIDTH * VIEWPORT_HEIGHT * 3];
  depthBuffer3D = new float[VIEWPORT_WIDTH * VIEWPORT_HEIGHT];

  // Habilita cull_face 
  glEnable(GL_CULL_FACE);

  // Seta modo de renderizacao
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  /* habilita a texturização */
  glEnable(GL_TEXTURE_3D);

  /* habilita composicao */
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  /* instala shaders */
  instaleShaders();
}

//============   Events Handling
void reshape(int x, int y)
{
  //VIEWPORT_WIDTH = x; VIEWPORT_HEIGHT = y;

  // Fixa-se as dimensoes por simplicidade. O mapa de profundidade
  // depende das dimensoes da janela
  glViewport(0, 0, VIEWPORT_WIDTH, VIEWPORT_HEIGHT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-0.2, 1.2, -0.2, 1.2, 0.0, 8.0); 

  glMatrixMode(GL_MODELVIEW);
}

// Arcball control is based on Ken Shoemakes from the Graphic Gems series
void MouseButton(int button, int state, int x, int y)
{
  GLint viewport[4];
  GLdouble modelview[16];
  GLdouble projection[16];
  float d, winX, winY, winZ;

  // Responde ao pressionamento nos botoes do mouse
  // Respond to mouse button presses.
  // If button1 pressed, mark this state so we know in motion function.

  if (button == GLUT_LEFT_BUTTON ) {
    g_bButton1Down = (state == GLUT_DOWN) ? 1 : 0;
    if (state == GLUT_DOWN) {
      // Compute o vetor unitário a partir do volCenter na direção do cursor 3D
      glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
      glGetDoublev( GL_PROJECTION_MATRIX, projection );
      glGetIntegerv( GL_VIEWPORT, viewport );

      winX = (float)x;
      winY = (float)viewport[3] - (float)y;
      winZ = depthBuffer3D[((int)winY)*viewport[2]+((int)winX)];

      gluUnProject( winX, winY, winZ, modelview, projection, viewport, 
		    &cursor[0],&cursor[1], &cursor[2]);

      winX = cursor[0]-volCenter[0];
      winY = cursor[1]-volCenter[1];
      winZ = cursor[2]-volCenter[2];

      d = sqrt(winX*winX + winY*winY + winZ*winZ);

      winX /= d;
      winY /= d;
      winZ /= d;

      prev_Pos[0] = winX;
      prev_Pos[1] = winY;
      prev_Pos[2] = winZ;
    } 
  } else if (button == GLUT_RIGHT_BUTTON) {
    g_bButton2Down = (state == GLUT_DOWN) ? 1 : 0;
    if (state == GLUT_DOWN) {
      prev_winX = x; prev_winY = y; 
    }
  } else if (button == 3) {  // Mouse wheel
    scale = (scale > 0.01)?scale - 0.01:scale;
    glutPostRedisplay();
  } else if (button == 4) {  // Mouse wheel
    scale = (scale < 5.0)?scale + 0.01:scale;
    glutPostRedisplay();
  }
}

// Multiplicao de 2 quaternios
void Qt_Mult (double *q1, double* q2, double *q)
{
  q[0] = q1[3]*q2[0]+q1[0]*q2[4]+q1[1]*q2[3]-q1[2]*q2[1];
  q[1] = q1[3]*q2[1]+q1[1]*q2[4]+q1[2]*q2[0]-q1[0]*q2[2];
  q[2] = q1[3]*q2[2]+q1[2]*q2[4]+q1[0]*q2[1]-q1[1]*q2[0];
  q[3] = q1[3]*q2[3]-q1[0]*q2[0]-q1[1]*q2[1]-q1[2]*q2[2];
}

void Qt2Matrix (double *q, float *mat) 
{
  double Nq = q[0]*q[0]+q[1]*q[1]+q[2]*q[2]+q[3]*q[3];

  mat[0] = 1.-(2.*(q[1]*q[1]+q[2]*q[2]))/Nq; 
  mat[1] = (2.*(q[0]*q[1]+q[2]*q[3]))/Nq;
  mat[2] = (2*(q[0]*q[2]-q[1]*q[3]))/Nq;
  mat[3] = 0.;
  mat[4] = (2.*(q[0]*q[1]-q[2]*q[3]))/Nq;
  mat[5] = 1.0-(2.*(q[0]*q[0]+q[2]*q[2]))/Nq;
  mat[6] = (2.*(q[1]*q[2]+q[0]*q[3]))/Nq;
  mat[7] = 0.;
  mat[8] = (2.*(q[0]*q[2]+q[1]*q[3]))/Nq;
  mat[9] = (2.*(q[1]*q[2]-q[0]*q[3]))/Nq;
  mat[10] = 1.0-(2.*(q[0]*q[0]+q[1]*q[1]))/Nq;
  mat[11] = mat[12] = mat[13] = mat[14] = 0.;
  mat[15] = 1.;
}

void reseta_accumMat() {
    accumMat[0] = accumMat[5] = accumMat[10] = accumMat[15] = 1.0;
    accumMat[1] = accumMat[2] = accumMat[3] = accumMat[4] = 
    accumMat[6] = accumMat[7] = accumMat[8] = accumMat[9] = 
    accumMat[11] = accumMat[12] = accumMat[13] = accumMat[14] = 0.; 

    prev_Qt[0] = prev_Qt[1] = prev_Qt[2] = 0.; prev_Qt[3] = 1.;
}

void MouseMotion(int x, int y)
{
  GLint viewport[4];
  GLdouble modelview[16];
  GLdouble projection[16];

  int dx, dy;
  float winX, winY, winZ, tmpX, tmpY, tmpZ, d;
  double quat[4], qatual[4], Nq;

  // Se botao direito apertado, move o cursor 3D
  // Origem do referencial do mouse é canto superior esquerdo
  if (g_bButton2Down) {
    dx = x - prev_winX;
    dy = y - prev_winY;
      
    float dz = sqrt(1.*(dx*dx + dy*dy))/(4*H); 
    if (y > prev_winY)  dz = -dz;

    tmpZ = winZ = prev_winZ + dz;

    glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
    glGetDoublev( GL_PROJECTION_MATRIX, projection );
    glGetIntegerv( GL_VIEWPORT, viewport );

    winX = (float)x;
    winY = (float)viewport[3] - (float)y;
 
    gluUnProject( winX, winY, winZ, modelview, projection, viewport, 
		  &cursor[0],&cursor[1], &cursor[2]);

  } else if ((g_bButton1Down) && (prev_winX != x || prev_winY != y)) {
    glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
    glGetDoublev( GL_PROJECTION_MATRIX, projection );
    glGetIntegerv( GL_VIEWPORT, viewport );

    tmpX = (float)x;
    tmpY = (float)viewport[3] - (float)y;
    tmpZ = depthBuffer3D[((int)tmpY)*viewport[2]+(int)tmpX];

    if (tmpZ == 1.0) return;

    gluUnProject( tmpX, tmpY, tmpZ, modelview, projection, viewport, 
		  &cursor[0],&cursor[1], &cursor[2]);

    winX = cursor[0]-volCenter[0];
    winY = cursor[1]-volCenter[1];
    winZ = cursor[2]-volCenter[2];

    d = sqrt(winX*winX + winY*winY + winZ*winZ);

    winX /= d;
    winY /= d;
    winZ /= d;

    //  win x prevPos
    quat[0] = prev_Pos[1]*winZ - prev_Pos[2]*winY;
    quat[1] = prev_Pos[2]*winX - prev_Pos[0]*winZ;
    quat[2] = prev_Pos[0]*winY - prev_Pos[1]*winX;
    quat[3] = prev_Pos[0]*winX+prev_Pos[1]*winY+prev_Pos[2]*winZ;

    Qt_Mult(quat, prev_Qt, qatual);

    Qt2Matrix(qatual, accumMat);

    // Atualiza para proxima iteracao
    prev_Pos[0] = winX; prev_Pos[1] = winY; prev_Pos[2] = winZ;
    memcpy(prev_Qt, qatual, sizeof(float)*4);
  } else {
    return;
  }
  prev_winX = x;
  prev_winY = y;
  prev_winZ = tmpZ;

  glutPostRedisplay();
}

void PassiveMotion(int x, int y)
{
  GLint viewport[4];
  GLdouble modelview[16];
  GLdouble projection[16];
 
  int dx, dy;
  float winX, winY, winZ;

  glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
  glGetDoublev( GL_PROJECTION_MATRIX, projection );
  glGetIntegerv( GL_VIEWPORT, viewport );

  winX = (float)x;
  winY = (float)viewport[3] - (float)y;
  winZ = prev_winZ;
 
  gluUnProject( winX, winY, winZ, modelview, projection, viewport, 
		&cursor[0], &cursor[1], &cursor[2]);

  prev_winX = x;
  prev_winY = y;

  glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
  //float v;

  switch (key) {
  case 'x':
    /* decrementa +5.0 no ângulo de rotação em torno do eixo x */
    spin_x -= 5.0;
    break;
  case 'X':
    /* incrementa +5.0 no ângulo de rotação em torno do eixo x */
    spin_x += 5.0;
    break;
  case 'y':
    /* decrementa +5.0 no ângulo de rotação em torno do eixo y */
    spin_y -= 5.0;
    break;
  case 'Y':
    /* incrementa +5.0 no ângulo de rotação em torno do eixo y */
    spin_y += 5.0;
    break;
  case 'z':
    /* decrementa +5.0 no ângulo de rotação em torno do eixo z */
    spin_z -= 5.0;
    break;
  case 'Z':
    /* incrementa +5.0 no ângulo de rotação em torno do eixo z */
    spin_z += 5.0;
    break;
  case 'r':
  case 'R':
    /* Reseta */
    spin_x = spin_y = spin_z = 0.0;
    glMatrixMode(GL_MODELVIEW);
    reseta_accumMat();
    glLoadIdentity();
    nVolSideZ = 0.5;
    noise = 0.25;
    alpha = 0.02;
    passo=0.005;
    prev_winZ=0.5;  
  break;
  case 'n':
    /* Diminui o limiar do ruido*/
    noise = (noise >= 0.01)?noise - 0.05:noise;
    break;
  case 'N':
    /* Aumenta o limiar do ruido */
    noise = (noise <= 1.0)?noise + 0.05:noise;
    break;
  case 'p':
    /* Diminui o passo */
    passo = (passo >= 0.0001)?passo - 0.002:passo;
    break;
  case 'P':
    /* Aumenta o passo */
    passo = (passo >= 0.2)?passo + 0.002:passo;
    break;
  case '+':
    /* retrocede o plano de corte */
    nVolSideZ = (nVolSideZ <= 1.0)? nVolSideZ + 0.01:nVolSideZ;
    break;
  case '-':
    /* avança o plano de corte */
    nVolSideZ = (nVolSideZ >= 0.0)? nVolSideZ - 0.01:nVolSideZ;
    break;
  case 27:
    /* libera memoria no cliente */
    delete [] data.buffer;
    /* libera memoria no servidor */
    glDeleteTextures (3, texID);
    glDeleteBuffers(3, vbo);
    exit(0);
    return;
  }

  /* Gera o evento "Display" */
  glutPostRedisplay();
}

int main(int argc, char *argv[])
{
  int c;
  std::string filename;

  /* Carrega a imagem 3D DICOM */
  if (argc == 2) {
    filename=argv[1];
    data.umax = -pow(2,30);
    data.umin = pow(2,30);
    if (!load.DICOMImage (filename, &data)) {
      std::cout << filename << " cannot be imported!" << std::endl;
      return 0;
    }
    std::cout << "Faixa de intensidades: " << data.umin << " - " << data.umax << std::endl;

  
  } else {
    std::cout << "Comando: ./jai13_glsl_dvr_int <caminho completo do volume de dados>" << std::endl;
    return 0;
  }

  /* Initializa OpenGL */
  glutInit(&argc, argv);

  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);

  /* Seta contexto GL */
  win = glutCreateWindow("Volume de Dados");

  // Verifca a versao de OpenGL: glxinfo | grep "OpenGL version"
  std::cout << "OpenGL version supported by this platform: " << glGetString(GL_VERSION) << std::endl;

  /* Outras inicializacoes */
#ifdef WIN32
  glewInit();
#endif
  init();

  /* Configura a janela */
  glutPositionWindow(10, 10);
  glutReshapeWindow(VIEWPORT_WIDTH, VIEWPORT_HEIGHT);

  H = (int)(sqrt(1.0*(VIEWPORT_WIDTH*VIEWPORT_WIDTH + VIEWPORT_HEIGHT*VIEWPORT_HEIGHT)));
  /* Define os tratadores de eventos */
  glutReshapeFunc(reshape);
  glutDisplayFunc(desenhe);
  glutKeyboardFunc(keyboard);
  glutMouseFunc (MouseButton);
  glutMotionFunc (MouseMotion);
  glutPassiveMotionFunc (PassiveMotion);

  /* Espera por eventos */
  glutMainLoop();
 
  return 0;
}
