/*  
 *  main_varray.cpp: Ray-casting based volume rendering 
 *                   for JAI 2013/CSBC mini-course (OpenGL 1.1)
 *
 *  Copyright (C) 2013  Wu Shin-Ting, FEEC, Unicamp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
//http://www.khronos.org/files/opengl-quick-reference-card.pdf
#ifdef WIN32
#include <windows.h>
#include <GL/glew.h>    
#else
#include <getopt.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <cstdio>
#include <cstring>
#include <string>
#include <iostream>
#include <stdio.h>     /* for printf */
#include <stdint.h>
#include <stdlib.h>    /* for exit */
#include <string.h>
#include <limits.h>
#include <math.h>
#include <GL/glut.h>    // Header File For The GLUT Library

#include "importDicom.h"
#include "transferFunction.h"

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

float vertices[][3] = {  // coordenadas espaciais dos vértices
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}, 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 0.0f}
};  

// Volume de textura: [1.0,1.0,1.0]
float texCoords0[][3] = {   // respectivas coordenadas de textura 
  {0.0f, 0.0f, 0.0f},
  {0.0f, 1.0f, 0.0f}, 
  {1.0f, 1.0f, 0.0f},
  {1.0f, 0.0f, 0.0f},

  {0.0f, 0.0f, 1.0f},
  {1.0f, 0.0f, 1.0f}, 
  {1.0f, 1.0f, 1.0f},
  {0.0f, 1.0f, 1.0f},

  {0.0f, 1.0f, 0.0f},
  {0.0f, 1.0f, 1.0f}, 
  {1.0f, 1.0f, 1.0f},
  {1.0f, 1.0f, 0.0f},

  {0.0f, 0.0f, 0.0f},
  {1.0f, 0.0f, 0.0f}, 
  {1.0f, 0.0f, 1.0f},
  {0.0f, 0.0f, 1.0f},

  {0.0f, 0.0f, 0.0f},
  {0.0f, 0.0f, 1.0f}, 
  {0.0f, 1.0f, 1.0f},
  {0.0f, 1.0f, 0.0f},

  {1.0f, 0.0f, 0.0f},
  {1.0f, 1.0f, 0.0f}, 
  {1.0f, 1.0f, 1.0f},
  {1.0f, 0.0f, 1.0f}
};

GLubyte quadIndices[] = {
    0,1,2,3,
    4,5,6,7,
    8,9,10,11,
    12,13,14,15,
    16,17,18,19,
    20,21,22,23
};

GLfloat spin_x=0.0, spin_y=0.0, spin_z=0.0;
GLfloat nVolSideZ=1.0;

Import load;
TransferFunction tfunc;

GLuint texID[2];
Import::ImgFormat data;
GLfloat nVolSides[3], scaleFactorVolSides[3], volCenter[3];
GLuint win, VIEWPORT_WIDTH=500, VIEWPORT_HEIGHT=500;

void transfereTexturas() {
  unsigned char *tf;
  int volSize, intensidade;
  unsigned char *texbuffer;

  // Gera nome de 1 texturas
  glGenTextures(1, texID);

  // Define a unidade de alinhamento nos acessos
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  // Constroi a funcao de transferencia (0=logaritmica; 1=linear; 2=quadratico; 3 cubico)
  int dim;
  tfunc.GetGrayScaleTF (1, data.umin, data.umax, &dim, &tf);

  volSize = data.dims[0]*data.dims[1]*data.dims[2];

  // Mapeia as intensidades em propriedades opticas (niveis de cinza)
  texbuffer = new unsigned char[volSize];
  memset(texbuffer, 0x00, volSize);

  int di, iz, iy, ix; 
  for (di=0, iz = 0; iz < data.dims[2]; iz++) {
    for (iy =0; iy < data.dims[1]; iy++) {
      for (ix =0; ix < data.dims[0]; ix++) {
	intensidade = (int)((reinterpret_cast<unsigned short*>(data.buffer))[iz*data.dims[0]*data.dims[1]+iy*data.dims[0]+ix]);
	if (intensidade)
	  texbuffer[di++] = (unsigned char)(tf[intensidade-data.umin+1]*2);
	else 
	  texbuffer[di++] = 0;
      }
    }
  }

  // Carrega o volume em niveis de cinza como textura 3D
  glBindTexture(GL_TEXTURE_3D, texID[0]);
  glTexImage3D(GL_TEXTURE_3D, 0, GL_LUMINANCE, data.dims[0],
  	       data.dims[1], data.dims[2], 0, GL_LUMINANCE,
  	       GL_UNSIGNED_BYTE, texbuffer);

  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

  // Libera memoria
  delete []texbuffer;
}

void crieVertexArray() {

  // Coordenadas espaciais dos vértices
  /* Face de fundo */
  vertices[1][1] = nVolSides[1]; 
  vertices[2][0] = nVolSides[0]; vertices[2][1] = nVolSides[1]; 
  vertices[3][0] = nVolSides[0]; 

  /* Face frontal */
  vertices[4][2] = nVolSides[2];
  vertices[5][0] = nVolSides[0]; vertices[5][2] = nVolSides[2];
  vertices[6][0] = nVolSides[0]; vertices[6][1] = nVolSides[1]; 
  vertices[6][2] = nVolSides[2];
  vertices[7][1] = nVolSides[1]; vertices[7][2] = nVolSides[2];

  /* Face de topo */
  vertices[8][1] = nVolSides[1]; 
  vertices[9][1] = nVolSides[1]; vertices[9][2] = nVolSides[2];
  vertices[10][0] = nVolSides[0]; vertices[10][1] = nVolSides[1]; 
  vertices[10][2] = nVolSides[2];
  vertices[11][0] = nVolSides[0]; vertices[11][1] = nVolSides[1]; 

  /* Face de apoio */
  vertices[13][0] = nVolSides[0]; 
  vertices[14][0] = nVolSides[0]; vertices[14][2] = nVolSides[2];
  vertices[15][2] = nVolSides[2];

  /* Face esquerda */
  vertices[17][2] = nVolSides[2];
  vertices[18][1] = nVolSides[1]; vertices[18][2] = nVolSides[2];
  vertices[19][1] = nVolSides[1];

  /* Face direita */
  vertices[20][0] = nVolSides[0]; 
  vertices[21][0] = nVolSides[0]; vertices[21][1] = nVolSides[1];
  vertices[22][0] = nVolSides[0]; vertices[22][1] = nVolSides[1]; 
  vertices[22][2] = nVolSides[2];
  vertices[23][0] = nVolSides[0]; vertices[23][2] = nVolSides[2];

  // Cria arranjos de dados no cliente
  glEnableClientState (GL_VERTEX_ARRAY);
  glVertexPointer (3, GL_FLOAT, 0, vertices);  // coord. compactadamente armazenadas
  glEnableClientState (GL_TEXTURE_COORD_ARRAY);
  glTexCoordPointer (3, GL_FLOAT, 0, texCoords0); // coord. compactamente armazenadas
}

void desenheProxyGeometry() 
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Seta os parametros de transformacao geometrica
  glLoadIdentity();
  glTranslatef(0., 0., -5.);
  glTranslatef(volCenter[0], volCenter[1], volCenter[2]);
  glRotatef(spin_x, 1.0, 0.0, 0.0);
  glRotatef(spin_y, 0.0, 1.0, 0.0);
  glRotatef(spin_z, 0.0, 0.0, 1.0);
  glTranslatef(-volCenter[0], -volCenter[1], -volCenter[2]);

  vertices[4][2] = nVolSideZ*nVolSides[2];
  vertices[5][2] = nVolSideZ*nVolSides[2];
  vertices[6][2] = nVolSideZ*nVolSides[2];
  vertices[7][2] = nVolSideZ*nVolSides[2];
  vertices[9][2] = nVolSideZ*nVolSides[2];
  vertices[10][2] = nVolSideZ*nVolSides[2];
  vertices[14][2] = nVolSideZ*nVolSides[2];
  vertices[15][2] = nVolSideZ*nVolSides[2];
  vertices[17][2] = nVolSideZ*nVolSides[2];
  vertices[18][2] = nVolSideZ*nVolSides[2];
  vertices[22][2] = nVolSideZ*nVolSides[2];
  vertices[23][2] = nVolSideZ*nVolSides[2];

  texCoords0[4][2] = nVolSideZ;
  texCoords0[5][2] = nVolSideZ;
  texCoords0[6][2] = nVolSideZ;
  texCoords0[7][2] = nVolSideZ;
  texCoords0[9][2] = nVolSideZ;
  texCoords0[10][2] = nVolSideZ;
  texCoords0[14][2] = nVolSideZ;
  texCoords0[15][2] = nVolSideZ;
  texCoords0[17][2] = nVolSideZ;
  texCoords0[18][2] = nVolSideZ;
  texCoords0[22][2] = nVolSideZ;
  texCoords0[23][2] = nVolSideZ;

  // "Compacta" os dados em arranjos no cliente para reduzir 
  // a quantidade de chamadas e, portanto, o desempenho 

  glDrawElements(GL_QUADS, 24, GL_UNSIGNED_BYTE, quadIndices); 

  //Esvazia os buffers  
  glFlush();
}

void init()
{
  // Define a cor em cinza (0.5,0.5,0.5) como cor de "CLEAR" */
  glClearColor (0.5, 0.5, 0.5, 1.0);

  // Transfere os dados como textura para GPU (servidor)
  transfereTexturas();

  // Normaliza as coordenadas em relação ao maior lado do volume
  float volSides[3], maxVolSide;
  int i;

  for (i = 0; i < 3; i++) {
    volSides[i] = data.dims[i] * data.space[i];
  }

  maxVolSide = MAX(MAX(volSides[0], volSides[1]), volSides[2]);

  for (i = 0; i < 3; i++) {
    scaleFactorVolSides[i] = maxVolSide/volSides[i];
    nVolSides[i] = volSides[i]/maxVolSide;
    volCenter[i] = nVolSides[i]/2.0;
  }

  // Cria vertex array no cliente
  crieVertexArray();

  // Habilita cull_face 
  glEnable(GL_CULL_FACE);

  // Seta modo de renderizacao
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  /* habilita a texturização */
  glEnable(GL_TEXTURE_3D);
}

void reshape(int x, int y)
{
    VIEWPORT_WIDTH = x; VIEWPORT_HEIGHT = y;

    glViewport(0, 0, VIEWPORT_WIDTH, VIEWPORT_HEIGHT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-0.5, 1.5, -0.5, 1.5, 0.0, 8.0); 

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void keyboard(unsigned char key, int x, int y)
{
  //float v;

  switch (key) {
  case 'x':
  case 'X':
    /* incrementa +5.0 no ângulo de rotação em torno do eixo x */
    spin_x += 5.0;
    break;
  case 'y':
  case 'Y':
    /* incrementa +5.0 no ângulo de rotação em torno do eixo y */
    spin_y += 5.0;
    break;
  case 'z':
  case 'Z':
   /* incrementa +5.0 no ângulo de rotação em torno do eixo z */
    spin_z += 5.0;
    break;
  case 'r':
  case 'R':
    /* Reseta */
    spin_x = spin_y = spin_z = 0.0;
    nVolSideZ = 1.0;
    break;
  case '+':
    /* retrocede o plano de corte */
    if (nVolSideZ + 0.05 < 1.0) nVolSideZ += 0.05;
    else nVolSideZ = 1.0;
    break;
  case '-':
    /* avança o plano de corte */
    if (nVolSideZ - 0.05 > 0.0) nVolSideZ -= 0.05;
    else nVolSideZ = 0.0;
    break;
  case 27:
    /* libera memoria no cliente */
    delete [] data.buffer;
    /* libera memoria no servidor */
    glDeleteTextures (1, texID);
    exit(0);
    return;
  }
  /* Gerar o evento "Display" */
  glutPostRedisplay();
}

int main(int argc, char *argv[])
{
  int c;
  std::string filename;

  /* Carrega a imagem 3D DICOM */
  if (argc == 2) {
    filename=argv[1];
    data.umax = -pow(2,30);
    data.umin = pow(2,30);
    if (!load.DICOMImage (filename, &data)) {
      std::cout << filename << " cannot be imported!" << std::endl;
      return 0;
    }
    std::cout << "Faixa de intensidades: " << data.umin << " - " << data.umax << std::endl;
  } else {
    std::cout << "Comando: ./jai13_varray <caminho completo do volume de dados>" << std::endl;
    return 0;
  }

  /* Inicializa OpenGL */
  glutInit(&argc, argv);

  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_SINGLE);

  /* Seta contexto GL */
  win = glutCreateWindow("Volume de Dados");

  // Verifca a versao de OpenGL: glxinfo | grep "OpenGL version"
  std::cout << "OpenGL version supported by this platform: " << glGetString(GL_VERSION) << std::endl;

  /* Outras inicializacoes */
#ifdef WIN32
  glewInit();
#endif
  init();

  /* Configura a janela */
  glutPositionWindow(10, 10);
  glutReshapeWindow(VIEWPORT_WIDTH, VIEWPORT_HEIGHT);

  /* Define os tratadores de eventos */
  glutReshapeFunc(reshape);
  glutDisplayFunc(desenheProxyGeometry);
  glutKeyboardFunc(keyboard);

  /* Espera por eventos */
  glutMainLoop();
 
  return 0;
}
