#version 150
/*  
 *  dvr_depth.frag: shader de fragmento para gerar mapa de profundidade
 *       aplicando ray-casting frente-para-trás com um limiar de 
 *       intensidade pre-definida <noise>
 *
 *  Copyright (C) 2013  Wu Shin-Ting, FEEC, Unicamp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
uniform mat4 viewMatrix, projMatrix;    //matriz de MODELVIEW
uniform mat4 viewMatrixInv; //matriz inversa de MODELVIEW
uniform vec2 fparam;      //passo de percurso, clip
uniform vec4 scaleFactors;//fatores de escala do espaco de textura em relacao
                            // ao espaco do volume normalizado
uniform sampler3D VOLUME;   //densidades da neuroimagem 3D
uniform sampler1D TRANSFERFUNCTION; //função de transferência
in vec3 texCoord;           //coordenadas de textura
out vec4 FragColor;         //cor do fragmento

void main(void){
 vec4 texpos, initPos, geomDir, geomPos;
 float t, scalar, noise; 

 mat4 mvi = viewMatrixInv;
 vec4 invScaleFactors = 1./scaleFactors;
 noise = 0.25;

 // ponto de entrada do volume
 initPos = vec4(texCoord, 1.0)*scaleFactors;

 // direção do raio paralelo ao eixo z 
 geomDir = normalize(mvi * vec4(0.0,0.0,-1.0,0.0)); 

 // inicia com initPos
 t = 0;
 do {
    geomPos = initPos + geomDir*t;          //próxima amostra ao longo do raio
    texpos = geomPos * invScaleFactors;     // coordenadas de textura
    scalar = texture(VOLUME, texpos.xyz).r;
    if (scalar > noise) {
      geomPos = projMatrix * viewMatrix * geomPos;
      geomPos.z = (geomPos.z+1)/2.0;      
      FragColor = vec4(geomPos.z, geomPos.z, geomPos.z, 1.0);
      return; 
    }
    t += fparam[0]; 
 } while (t < 10.0); 
 
 FragColor = vec4(1.0, 1.0, 1.0, 1.0); 
}
