#version 150
/*  
 *  dvr_ftob.frag: shader de fragmento para gerar ray-casting
 *       frente-para-trás
 *
 *  Copyright (C) 2013  Wu Shin-Ting, FEEC, Unicamp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

uniform mat4 viewMatrixInv; //matriz inversa de MODELVIEW
uniform vec2 fparam;      //passo de percurso, clip
uniform vec4 scaleFactors;//fatores de escala do espaco de textura em relacao
                            // ao espaco do volume normalizado
uniform sampler3D VOLUME;   //densidades da neuroimagem 3D
uniform sampler1D TRANSFERFUNCTION; //função de transferência
in vec3 texCoord;           //coordenadas de textura
out vec4 FragColor;         //cor do fragmento

void main(void){
 vec4 texpos, initPos, geomDir, geomPos;
 float src, dst, t, scalar, alpha, controle; 

 mat4 mvi = viewMatrixInv;
 vec4 invScaleFactors = 1./scaleFactors;
 dst = src = 0.0;
 controle = 0.02;

 // ponto de entrada do volume
 initPos = vec4(texCoord, 1.0)*scaleFactors;

 // direção do raio paralelo ao eixo z 
 geomDir = normalize(mvi * vec4(0.0,0.0,-1.0,0.0)); 

 // inicia com initPos
 geomPos = initPos;          //primeira amostra ao longo do raio
 texpos = geomPos * invScaleFactors;     // coordenadas de textura
 scalar = texture(VOLUME, texpos.xyz).r;
 src = texture(TRANSFERFUNCTION, scalar).r;
 dst = controle*src;
 alpha = controle*src;
 t = fparam[0];
 do {
    geomPos = initPos + geomDir*t;          //próxima amostra ao longo do raio
    texpos = geomPos * invScaleFactors;     // coordenadas de textura
    if (texpos.x >= 0.0) {
    if (texpos.y >= 0.0) {
    if (texpos.z >= 0.0) {
    if (texpos.x <= 1.0) {
    if (texpos.y <= 1.0) {
    if (texpos.z <= 1.0) {
       scalar = texture(VOLUME, texpos.xyz).r;
       src = texture(TRANSFERFUNCTION, scalar).r;
       clamp (src, 0.0, 1.0);
       dst += (1-alpha)*(controle*src);     
       clamp (dst, 0.0, 1.0);
       alpha = (controle*src) + alpha*(1 - (controle*src)); 
       if (alpha >= 1.0) // 1a. condicao de parada: satura a opacidade
       	  break;
    }
    }
    }
    }
    }
}
    t += fparam[0]; 
 } while (t < 10.0); 
 
 if (alpha == 0.0)
    FragColor = vec4(0.5, 0.5, 0.5, 1.0); 
 else
    FragColor = vec4(dst, dst, dst, 1.0);  // escrever a cor final do fragmento
}
