#version 150
/*  
 *  depth_int.frag: shader de fragmento para renderizar somente a amostra
 *              visivel sem composicao
 *
 *  Copyright (C) 2013  Wu Shin-Ting, FEEC, Unicamp
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
uniform mat4 viewMatrix, projMatrix;    //matriz de MODELVIEW
uniform mat4 viewMatrixInv; //matriz inversa de MODELVIEW
uniform vec4 fparam;      //passo de percurso, clip, alpha, noise
uniform vec4 scaleFactors;//fatores de escala do espaco de textura em relacao
                            // ao espaco do volume normalizado
uniform sampler3D VOLUME;   //densidades da neuroimagem 3D
uniform sampler2DRect DEPTHMAP;   // mapa de profundidade
uniform sampler1D TRANSFERFUNCTION; //função de transferência
in vec3 texCoord;           //coordenadas de textura
out vec4 FragColor;         //cor do fragmento

void main(void){
 vec4 texpos, initPos, geomDir, geomPos, clipPos;
 float src, dst, t, scalar, alpha, clip, controle, noise;
 bool flag; 

//alpha = texture(DEPTHMAP, gl_FragCoord.xy).x;
//FragColor = vec4(alpha, alpha, alpha, 1.0);

//return;

 mat4 mv = viewMatrix;
 mat4 mvi = viewMatrixInv;
 vec4 invScaleFactors = 1./scaleFactors;
 alpha = dst = src = 0.0;
 controle = fparam[2];
 noise = fparam[3];
 clip = fparam[1];

 // ponto de entrada do volume
 initPos = vec4(texCoord, 1.0)*scaleFactors;

 // direção do raio paralelo ao eixo z 
 geomDir = normalize(mvi * vec4(0.0,0.0,-1.0,0.0)); 

 // inicia com initPos
 t = 0;
 flag = false;
 do {
    geomPos = initPos + geomDir*t;          //próxima amostra ao longo do raio
    clipPos = mv*geomPos;
    if (clipPos.z < clip+mv[3][2]) {
       texpos = geomPos * invScaleFactors;     // coordenadas de textura
       scalar = texture(VOLUME, texpos.xyz).r;
        if (scalar > noise) {
       	 src = texture(TRANSFERFUNCTION, scalar).r;
	 FragColor = vec4(src, src, src, 1.0);
	 return;
	 }
    }
    t += fparam[0]; 
 } while (t < 2.0); 
 
 if (alpha == 0.0)
    FragColor = vec4(0.5, 0.5, 0.5, 1.0); 
 else
    FragColor = vec4(dst, dst, dst, 1.0);  // escrever a cor final do fragmento
}
